"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.plugin = void 0;
var _typescript = require("@graphql-codegen/typescript");
var _typescriptResolvers = require("@graphql-codegen/typescript-resolvers");
var _pluginHelpers = require("@graphql-codegen/plugin-helpers");
var _graphql = require("graphql");
var _codemod = require("../codemod");
const plugin = async (schema, documents, config) => {
  const astNode = (0, _pluginHelpers.getCachedDocumentNodeFromSchema)(schema);
  function getInterface(interfaceName) {
    const interfaceNode = astNode.definitions.filter(n => n.kind === _graphql.Kind.INTERFACE_TYPE_DEFINITION).find(n => n.name.value === interfaceName);
    if (!interfaceNode) throw new Error(`No interface with name "${interfaceName}"`);
    return interfaceNode;
  }
  const entities = {};
  const propertyMapping = {};

  /**
   * Modify entities to have all fields optional except for keys
   */
  function handleEntityDirective(node) {
    if (!getDirective(node, "entity")) {
      return node;
    }
    entities[node.name.value] = {};
    // @ts-expect-error mutating
    node.fields = node.fields?.map(field => {
      if (getDirective(field, "key")) {
        entities[node.name.value][field.name.value] = getTypeName(field.type);
        return field;
      }
      if (field.type.kind === _graphql.Kind.NON_NULL_TYPE) {
        // @ts-expect-error mutating
        field.type = field.type.type;
      }
      return field;
    });
    return node;
  }
  const visitorEntityKey = {
    InterfaceTypeDefinition: {
      leave: node => {
        return handleEntityDirective(node);
      }
    },
    ObjectTypeDefinition: {
      leave: node => {
        const resultNode = handleEntityDirective(node);

        // If the object type implements an interface that is an entity, the
        // object is also an entity and should have the same keys
        resultNode.interfaces?.forEach(interfaceItem => {
          const interfaceNode = getInterface(interfaceItem.name.value);
          const directive = getDirective(interfaceNode, "entity");
          if (!directive) {
            return;
          }

          // @ts-expect-error mutating
          node.fields = node.fields?.map(field => {
            if (interfaceNode.fields?.find(f => f.name.value === field.name.value && getDirective(f, "key"))) {
              return field;
            }
            if (field.type.kind === _graphql.Kind.NON_NULL_TYPE) {
              // @ts-expect-error mutating
              field.type = field.type.type;
            }
            return field;
          });
        });
      }
    }
  };
  const vistorEntityProperty = {
    ObjectTypeDefinition: {
      leave: node => {
        const directive = getDirective(node, "entityProperty");
        if (!directive) {
          return node;
        }
        const entityArg = directive.arguments?.[0];
        if (entityArg?.value.kind === _graphql.Kind.STRING) {
          propertyMapping[node.name.value] = entityArg.value.value;
        }
        ;
        node.fields = node.fields?.map(field => {
          if (field.type.kind === _graphql.Kind.NON_NULL_TYPE) {
            ;
            field.type = field.type.type;
          }
          return field;
        });
        return node;
      }
    }
  };
  const result = (0, _graphql.visit)((0, _graphql.visit)(astNode, visitorEntityKey), vistorEntityProperty);
  const newSchema = (0, _graphql.buildASTSchema)(result);
  const tsOutput = await (0, _typescript.plugin)(newSchema, documents, config);
  const resolverOutput = await (0, _typescriptResolvers.plugin)(newSchema, documents, config);
  return {
    prepend: [...(tsOutput.prepend ?? []), ...(resolverOutput.prepend ?? [])],
    append: [...(tsOutput.append ?? []), ...(resolverOutput.append ?? [])],
    content: [(0, _codemod.codemod)(tsOutput.content, {
      entities,
      propertyMapping
    }), resolverOutput.content].join("\n")
  };
};
exports.plugin = plugin;
function getDirective(node, directiveName) {
  return node.directives?.find(directive => directive.name.value === directiveName);
}
function getTypeName(type) {
  return type.kind === _graphql.Kind.LIST_TYPE || type.kind === _graphql.Kind.NON_NULL_TYPE ? getTypeName(type.type) : type.name.value;
}