import core, { FileInfo, API } from "jscodeshift"
import { runPrettier } from "./utils"
import * as K from "ast-types/gen/kinds"

const ENTITY_SUFFIX = "__Entity"
const ENTITY_PROPERTY = "_entity"

// export type GraphQLType = "ID" | "String" | "Int" | "Float"
export type GraphQLType = string

export type EntityConfig = Record<string, GraphQLType>

export type Options = {
  entities: Record<string, EntityConfig>
  propertyMapping: Record<string, string>
}

export default function transform(
  file: FileInfo,
  { jscodeshift: j }: API,
  { entities, propertyMapping }: Options = {
    entities: {},
    propertyMapping: {},
  },
): string {
  Object.entries(propertyMapping).forEach(([property, entity]) => {
    if (!entities[entity]) {
      throw new Error(
        `Type ${property} defines @entityProperty("${entity}") but ${entity} does not exist`,
      )
    }
  })

  const source = j(file.source)

  function graphqlTypeToTSType(graphqlType: string) {
    const mapping: Record<string, K.TSTypeKind> = {
      ID: j.tsStringKeyword(),
      String: j.tsStringKeyword(),
      Int: j.tsNumberKeyword(),
      Float: j.tsNumberKeyword(),
      Boolean: j.tsBooleanKeyword(),
    }

    return mapping[graphqlType]
  }

  function addEntity(entityName: string, config: EntityConfig) {
    const result = source.find(j.TSTypeAliasDeclaration, {
      id: {
        type: "Identifier",
        name: `${entityName}${ENTITY_SUFFIX}`,
      },
    })

    // This entity has already been added
    if (result.length > 0) {
      return
    }

    insertAfterImports(
      j.tsTypeAliasDeclaration(
        j.identifier(`${entityName}${ENTITY_SUFFIX}`),
        j.tsTypeLiteral(
          Object.entries(config).map(([propertyName, type]) =>
            j.tsPropertySignature(
              j.identifier(propertyName),
              j.tsTypeAnnotation(graphqlTypeToTSType(type)),
            ),
          ),
        ),
      ),
    )
  }

  function insertAfterImports(insert: core.Node) {
    const imports = source.find(j.ImportDeclaration)
    const n = imports.length

    if (n) {
      //j(imports.at(0).get()).insertBefore(s); // before the imports
      j(imports.at(n - 1).get()).insertAfter(insert) // after the imports
    } else {
      source.get().node.program.body.unshift(insert) // beginning of file
    }
  }

  Object.entries(propertyMapping).forEach(([property, entity]) => {
    source
      .find(j.TSTypeAliasDeclaration, {
        id: {
          type: "Identifier",
          name: property,
        },
        typeAnnotation: {
          type: "TSTypeLiteral",
        },
      })
      .forEach((path) => {
        const typeLiteral = path.node.typeAnnotation
        if (typeLiteral.type === "TSTypeLiteral") {
          // Add entity type which can be referenced
          addEntity(entity, entities[entity])

          // Add property referencing the entity type
          typeLiteral.members.unshift(
            j.tsPropertySignature(
              j.identifier(ENTITY_PROPERTY),
              j.tsTypeAnnotation(
                j.tsTypeReference(j.identifier(`${entity}${ENTITY_SUFFIX}`)),
              ),
            ),
          )
        }
      })
  })

  return runPrettier(source.toSource())
}
