import "@graphql-codegen/testing"
import { buildSchema } from "graphql"
import { plugin } from "./plugin"
import { runPrettier } from "../codemod/utils"

const defaultOptions = {
  useTypeImports: true,
  skipTypename: true,
  strictScalars: true,
}

const directives = /* GraphQL */ `
  directive @entity on OBJECT | INTERFACE
  directive @entityProperty(entity: String!) on OBJECT
  directive @key on FIELD_DEFINITION
`

describe("plugin", () => {
  it("makes non key fields optional on entities", async () => {
    const schema = buildSchema(/* GraphQL */ `
      ${directives}

      type User @entity {
        id: ID! @key
        otherId: ID! @key
        age: Int!
        hobby: String
      }
    `)

    const result = await plugin(schema, [], defaultOptions)

    expect(result.content).toBeSimilarStringTo(
      runPrettier(`
        export type User = {
          id: Scalars['ID']['output'];
          otherId: Scalars['ID']['output'];
          age?: Maybe<Scalars['Int']['output']>;
          hobby?: Maybe<Scalars['String']['output']>;
        };
      `),
    )
  })

  it("no changes without directives", async () => {
    const schema = buildSchema(/* GraphQL */ `
      type User {
        id: ID!
        otherId: ID!
        age: Int!
        hobby: String
      }
    `)

    const result = await plugin(schema, [], defaultOptions)

    expect(result.content).toBeSimilarStringTo(
      runPrettier(`
        export type User = {
          id: Scalars['ID']['output'];
          otherId: Scalars['ID']['output'];
          age: Scalars['Int']['output'];
          hobby?: Maybe<Scalars['String']['output']>;
        };
      `),
    )
  })

  it("no changes without directives", async () => {
    const schema = buildSchema(/* GraphQL */ `
      ${directives}

      type User @entity {
        id: ID! @key
        otherId: ID! @key
        age: Int!
        hobby: String
        name: UserNames!
        location: Point
      }

      type UserNames @entityProperty(entity: "User") {
        first: String!
        last: String!
      }

      type Point {
        lat: Float!
        lng: Float!
      }
    `)

    const result = await plugin(schema, [], defaultOptions)

    expect(result.content).toBeSimilarStringTo(
      runPrettier(`
        export type User = {
          id: Scalars["ID"]["output"]
          otherId: Scalars["ID"]["output"]
          age?: Maybe<Scalars["Int"]["output"]>
          hobby?: Maybe<Scalars["String"]["output"]>
          name?: Maybe<UserNames>
          location?: Maybe<Point>
        };
      `),
    )

    expect(result.content).toBeSimilarStringTo(
      runPrettier(`
        export type UserNames = {
          _entity: User__Entity
          first?: Maybe<Scalars["String"]["output"]>
          last?: Maybe<Scalars["String"]["output"]>
        };
      `),
    )

    expect(result.content).toBeSimilarStringTo(
      runPrettier(`
        export type Point = {
          lat: Scalars["Float"]["output"]
          lng: Scalars["Float"]["output"]
        };
      `),
    )
  })

  it("works with interfaces", async () => {
    const schema = buildSchema(/* GraphQL */ `
      ${directives}

      interface Post @entity {
        id: ID! @key
        description: String!
      }

      type PlainTextPost implements Post {
        id: ID!
        description: String!
        plainTextStuff: String!
      }

      type ImagePost implements Post {
        id: ID!
        description: String!
        imageStuff: String!
      }
    `)

    const result = await plugin(schema, [], defaultOptions)

    expect(result.content).toBeSimilarStringTo(
      runPrettier(`
        export type Post = {
          id: Scalars['ID']['output'];
          description?: Maybe<Scalars['String']['output']>;
        };
      `),
    )

    expect(result.content).toBeSimilarStringTo(
      runPrettier(`
        export type PlainTextPost = Post & {
          id: Scalars["ID"]["output"]
          description?: Maybe<Scalars["String"]["output"]>
          plainTextStuff?: Maybe<Scalars["String"]["output"]>
        };
      `),
    )

    expect(result.content).toBeSimilarStringTo(
      runPrettier(`
        export type ImagePost = Post & {
          id: Scalars["ID"]["output"]
          description?: Maybe<Scalars["String"]["output"]>
          imageStuff?: Maybe<Scalars["String"]["output"]>
        };
      `),
    )
  })

  it("throw error if an @entityProperty specifies a non existent entity", async () => {
    const schema = buildSchema(/* GraphQL */ `
      ${directives}

      type Person @entity {
        id: Int! @key
        details: PersonDetails
      }

      type PersonDetails @entityProperty(entity: "NotPerson") {
        name: String!
      }
    `)

    await expect(
      plugin(schema, [], defaultOptions),
    ).rejects.toMatchInlineSnapshot(
      `[Error: Type PersonDetails defines @entityProperty("NotPerson") but NotPerson does not exist]`,
    )
  })
})
