import {
  plugin as tsPlugin,
  TypeScriptPluginConfig,
} from "@graphql-codegen/typescript"
import { plugin as resolverPlugin } from "@graphql-codegen/typescript-resolvers"
import {
  PluginFunction,
  Types,
  getCachedDocumentNodeFromSchema,
} from "@graphql-codegen/plugin-helpers"
import {
  GraphQLSchema,
  visit,
  buildASTSchema,
  ASTVisitor,
  Kind,
  FieldDefinitionNode,
  ObjectTypeDefinitionNode,
  InterfaceTypeDefinitionNode,
  TypeNode,
} from "graphql"
import { codemod, Options } from "../codemod"

export const plugin: PluginFunction<
  TypeScriptPluginConfig,
  Types.ComplexPluginOutput
> = async (
  schema: GraphQLSchema,
  documents: Types.DocumentFile[],
  config: TypeScriptPluginConfig,
) => {
  const astNode = getCachedDocumentNodeFromSchema(schema)

  function getInterface(interfaceName: string): InterfaceTypeDefinitionNode {
    const interfaceNode = astNode.definitions
      .filter(
        (n): n is InterfaceTypeDefinitionNode =>
          n.kind === Kind.INTERFACE_TYPE_DEFINITION,
      )
      .find((n) => n.name.value === interfaceName)

    if (!interfaceNode)
      throw new Error(`No interface with name "${interfaceName}"`)

    return interfaceNode
  }

  const entities: Options["entities"] = {}
  const propertyMapping: Options["propertyMapping"] = {}

  /**
   * Modify entities to have all fields optional except for keys
   */
  function handleEntityDirective(
    node: ObjectTypeDefinitionNode | InterfaceTypeDefinitionNode,
  ) {
    if (!getDirective(node, "entity")) {
      return node
    }

    entities[node.name.value] = {}
    // @ts-expect-error mutating
    node.fields = node.fields?.map((field) => {
      if (getDirective(field, "key")) {
        entities[node.name.value][field.name.value] = getTypeName(field.type)
        return field
      }

      if (field.type.kind === Kind.NON_NULL_TYPE) {
        // @ts-expect-error mutating
        field.type = field.type.type
      }

      return field
    })

    return node
  }

  const visitorEntityKey: ASTVisitor = {
    InterfaceTypeDefinition: {
      leave: (node) => {
        return handleEntityDirective(node)
      },
    },

    ObjectTypeDefinition: {
      leave: (node) => {
        const resultNode = handleEntityDirective(node)

        // If the object type implements an interface that is an entity, the
        // object is also an entity and should have the same keys
        resultNode.interfaces?.forEach((interfaceItem) => {
          const interfaceNode = getInterface(interfaceItem.name.value)
          const directive = getDirective(interfaceNode, "entity")

          if (!directive) {
            return
          }

          // @ts-expect-error mutating
          node.fields = node.fields?.map((field) => {
            if (
              interfaceNode.fields?.find(
                (f) =>
                  f.name.value === field.name.value && getDirective(f, "key"),
              )
            ) {
              return field
            }

            if (field.type.kind === Kind.NON_NULL_TYPE) {
              // @ts-expect-error mutating
              field.type = field.type.type
            }

            return field
          })
        })
      },
    },
  }

  const vistorEntityProperty: ASTVisitor = {
    ObjectTypeDefinition: {
      leave: (node) => {
        const directive = getDirective(node, "entityProperty")

        if (!directive) {
          return node
        }

        const entityArg = directive.arguments?.[0]

        if (entityArg?.value.kind === Kind.STRING) {
          propertyMapping[node.name.value] = entityArg.value.value
        }

        ;(node.fields! as any) = node.fields?.map((field) => {
          if (field.type.kind === Kind.NON_NULL_TYPE) {
            ;(field.type as any) = field.type.type
          }

          return field
        })

        return node
      },
    },
  }

  const result = visit(visit(astNode, visitorEntityKey), vistorEntityProperty)
  const newSchema = buildASTSchema(result)

  const tsOutput = await tsPlugin(newSchema, documents, config)
  const resolverOutput = await resolverPlugin(newSchema, documents, config)

  return {
    prepend: [...(tsOutput.prepend ?? []), ...(resolverOutput.prepend ?? [])],
    append: [...(tsOutput.append ?? []), ...(resolverOutput.append ?? [])],
    content: [
      codemod(tsOutput.content, { entities, propertyMapping }),
      resolverOutput.content,
    ].join("\n"),
  }
}

function getDirective(
  node:
    | FieldDefinitionNode
    | ObjectTypeDefinitionNode
    | InterfaceTypeDefinitionNode,
  directiveName: string,
) {
  return node.directives?.find(
    (directive) => directive.name.value === directiveName,
  )
}

function getTypeName(type: TypeNode): string {
  return type.kind === Kind.LIST_TYPE || type.kind === Kind.NON_NULL_TYPE
    ? getTypeName(type.type)
    : type.name.value
}
